
#ifndef __THREAD_H
    #define __THREAD_H

    #ifdef __unix
        #ifndef __unix__
            #define __unix__
        #endif
    #endif
    #ifndef __unix__
        #ifdef __APPLE__ & __MACH__
            #define __unix__
        #endif
    #endif

    #ifdef __unix__
        #include <pthread.h>
    #elif __WIN32__ || _MSC_VER || _Windows || __NT__
        #ifndef __windows__
            #define __windows__
        #endif
        #include <windows.h>
        #include <process.h>
        #include <stdint.h>
    #else
        #error "No multithreading mode found for target operating system"
    #endif

    class Mutex
    {
        protected:
            #ifdef __unix__
                pthread_mutex_t mutex;
            #elif __windows__
                HANDLE mutex;
            #endif
        public:
            Mutex();
            ~Mutex();
            
            void Lock();
            bool TryLock();
            void Unlock();
    };
    
    class ConditionVariable : public Mutex
    {
        protected:
            #ifdef __unix__
                pthread_cond_t cond;
            #elif __windows__
                CONDITION_VARIABLE cond;
            #endif
        public:
            ConditionVariable();
            ~ConditionVariable();

            bool Wait();
            void SignalOne();
            void SignalAll();
    };    

    
    #ifdef __unix__
        void* ThreadWrapper(void*);
    #elif __windows__
        void ThreadWrapper(void*);
    #endif
    class Thread
    {
        #ifdef __unix__
            friend void* ThreadWrapper(void*);
        #elif __windows__
            friend void ThreadWrapper(void*);
        #endif
        protected:
            Mutex mutex;
            bool Started;
            #ifdef __unix__
                pthread_t thread;
            #elif __windows__
                uintptr_t thread;
            #endif

            virtual void RunThread() = 0;
        public:
            void Start();
            void Terminate();
            static void TestTermination();
            void Join();
            
            void Lock();
            void Unlock();
            bool TryLock();
            Thread();
            virtual ~Thread();
    };

#endif